/*             CSE 687 Project 2 --- String Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * string.cc ---- the module of the member & frind functions       *
 *                for class String and its derived class BCstring  *
 *******************************************************************/

#include "string.hh"

extern void error(char *);      // error handler

///////////// class String member functions ///////////////

// ----<String :: String(void)>---- //
// - copy constructs String object (a null string)
String :: String(void)
{
  slen = 0;
  if ((shead = (sptr = new char[1])) == '\0') 
    error("error: exhaust system memory\n"); 
  *shead = '\0';
}

// ----<String :: String(const String& s)>---- //
// - copy constructs String object from an existing object
String :: String(const String& s)
{
  slen = s.slen;
  if ((shead = (sptr = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n"); 
  strcpy(shead, s.shead);
}

// ----<String :: String(const char *s)>---- //
// - constructs String object by promoting char* object
String :: String(const char *s)
{
  if (s == '\0') error("error: invalid string\n"); // check invalid string  
  slen = strlen(s);
  if ((shead = (sptr = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n"); 
  strcpy(shead, s);  
}

// ----<String :: String(const int n)>---- //
// - allocate space for a String with a length n and 
//   blank char is filled in the new string
String :: String(const int n)
{
  if (n<0) error("invalid string length\n");

  slen = n;
  if ((shead = (sptr = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n");
  for (int i = 0; i<slen; shead[i++] = ' ');
  shead[slen] = '\0';   
}

// ----<String& String :: operator= (const String& str)>---- //
// - sets String object equal to object str
String& String :: operator= (const String& str)
{ 
  if (*this == str) return *this; // for c1 = c1 case
  if (shead) delete []shead;
  slen = str.slen;
  if ((sptr = (shead = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n"); 
  strcpy(shead,str.shead);  // copy str's data
  
  return *this;
}

// ----<String operator+ (String& s1, String& s2)>---- //
// - concatenation of two strings
String operator+ (String& s1, String& s2)
{
  int length = s1.slen + s2.slen;
  char *tptr;
  if ((tptr = new char[length+1]) == '\0') 
      error("error: exhaust system memory\n"); 

  strcpy(tptr,s1.shead);    // copy s1's data
  strcat(tptr,s2.shead);    // copy s2's data
  String temp(tptr);  // use togather with unary operator = which
                      // copies this temporary immediately exit 
                      //from this function

  return temp;
}

// ----<istream& operator>> (istream& istrm, String& str)>---- //
// - overloading istream operator >>
istream& operator>> (istream& istrm, String& str)
{
  #define MAXSIZE 4095
  char *buffer;

  if ((buffer = new char[MAXSIZE+1]) == '\0') 
      error("error: exhaust system memory\n"); 
  istrm.getline(buffer, MAXSIZE);
  buffer[MAXSIZE] = '\0';
  str.slen = strlen(buffer);
  if (str.shead) delete str.shead;
  if ((str.shead = (str.sptr = new char[str.slen+1])) == '\0')
      error("error: exhaust system memory\n"); 
  strcpy (str.shead, buffer);
  delete []buffer;
  return istrm;

  #undef MAXSIZE
}

///////////// class BCstring member functions ///////////////

// ----<char& BCstring :: operator[] (const int n)>---- //
// - return ing reference allows writing and reading indexed data
char& BCstring :: operator[] (const int n)
{    
  if (n < 0 || n > strlen(shead)+1 ) {
    error ("error: index outside string\n");};
  sptr = shead + n;

  return shead[n];
}

// ----<BCstring& BCstring :: operator= (const BCstring& str)>---- //
// - sets String object equal to object str
BCstring& BCstring :: operator= (const BCstring& str)
{ 
  if (*this == str) return *this; // for c1 = c1 case

  // if source string has the same length as target string
  // directly perform copy
  if (slen != str.slen) { // not the same
    if (shead) delete []shead;
    slen = str.slen;
    if ((sptr = (shead = new char[slen+1])) == '\0') 
      error("error: exhaust system memory\n"); 
  }; // otherwise, use the previous space
  strcpy(shead,str.shead);  // copy str's data
  
  return *this;
}

// ----<BCstring& BCstring :: operator= (const String& str)>---- //
// - sets String object equal to object str
BCstring& BCstring :: operator= (const String& str)

{ 
  if (*this == str) return *this; // for c1 = c1 case

  // if source string has the same length as target string
  // directly perform copy
  if (slen != str.slen) { // not the same
    if (shead) delete []shead;
    slen = str.slen;
    if ((sptr = (shead = new char[slen+1])) == '\0') 
      error("error: exhaust system memory\n"); 
  }; // otherwise, use the previous space
  strcpy(shead,str.shead);  // copy str's data
  
  return *this;
}

// ----<BCstring operator+ (BCstring& s1, BCstring& s2)>---- //
// - concatenation of two strings
BCstring operator+ (BCstring& s1, BCstring& s2)
{
  int ltmp,length = strlen(s1.shead);

  if (s1.slen != length || s2.slen != (ltmp = strlen(s2.shead))) {
    cerr << "warning: incorrect string length\n";
    length += ltmp;
  }
  else 
    length = s1.slen + s2.slen;

  char *tptr;
  if ((tptr = new char[length+1]) == '\0') 
      error("error: exhaust system memory\n"); 

  strcpy(tptr,s1.shead);    // copy s1's data
  strcat(tptr,s2.shead);    // copy s2's data
  String temp(tptr);  // use togather with unary operator = which
                      // copies this temporary immediately exit 
                      //from this function

  return temp;
}


